/*
 *======================================================================
 *
 *Copyright 1992 Sun Microsystems, Inc.
 *
 *The Interface Definition Language Compiler Front End (CFE) is made
 *available for use provided that this legend is included on all media and
 *documentation and as a part of the software program in whole or part.
 *Users may copy and extend functionality (but may not remove
 *functionality) of the Interface Definition Language CFE without charge,
 *but are not authorized to license or distribute it to anyone else except
 *as part of a product or program developed by the user or with the express
 *written consent of Sun Microsystems, Inc.
 *
 *The names of Sun Microsystems, Inc. and any of its subsidiaries may not
 *be used in advertising or publicity pertaining to distribution of
 *Interface Definition Language CFE as permitted herein.
 *
 *The Interface Definition Language CFE may not be exported outside the
 *United States without first obtaining the appropriate government
 *approvals.
 *
 *INTERFACE DEFINITION LANGUAGE CFE IS PROVIDED AS IS WITH NO WARRANTIES
 *OF ANY KIND INCLUDING THE WARRANTIES OF DESIGN, MERCHANTIBILITY AND
 *FITNESS FOR A PARTICULAR PURPOSE, NONINFRINGEMENT, OR ARISING FROM A
 *COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 *
 *Interface Definition Language CFE is provided with no support and
 *without any obligation on the part of Sun Microsystems, Inc. or any of
 *its subsidiaries or affiliates to assist in its use, correction,
 *modification or enhancement.
 *
 *SUN MICROSYSTEMS, INC. OR ANY OF ITS SUBSIDIARIES OR AFFILIATES SHALL
 *HAVE NO LIABILITY WITH RESPECT TO THE INFRINGEMENT OF COPYRIGHTS, TRADE
 *SECRETS OR ANY PATENTS BY INTERFACE DEFINITION LANGUAGE CFE OR ANY PART
 *THEREOF.
 *
 *IN NO EVENT WILL SUN MICROSYSTEMS, INC. OR ANY OF ITS SUBSIDIARIES OR
 *AFFILIATES BE LIABLE FOR ANY LOST REVENUE OR PROFITS OR OTHER SPECIAL,
 *INDIRECT AND CONSEQUENTIAL DAMAGES, EVEN IF SUN HAS BEEN ADVISED OF THE
 *POSSIBILITY OF SUCH DAMAGES.
 *
 *SunSoft, Inc.  
 *2550 Garcia Avenue 
 *Mountain View, California  94043
 *
 *
 *
 *
 *======================================================================
 */

#pragma ident "%@(#)ast_attribute.cc	1.46% %92/06/11% Sun Microsystems"

/*
 * ast_attribute.cc - Implementation of class AST_Attribute.cc
 *
 * AST_Attribute nodes denote IDL attribute declarations.
 * AST_Attribute nodes are AST_Fields with a readonly indication.
 * Hence they have a name (an UTL_ScopedName), a type (a subtype
 * of AST_Type) and a boolean indicating whether the attribute is
 * readonly.
 */

#include	<idl.hh>
#include	<idl_extern.hh>

/*
 * Constructor(s) and destructor
 */
AST_Attribute::AST_Attribute()
	     : pd_readonly(TRUE)
{
}

AST_Attribute::AST_Attribute(boolean ro,
			     AST_Type *ft,
			     UTL_ScopedName *n,
			     UTL_StrList *p)
	     : AST_Field(AST_Decl::NT_attr, ft, n, p),
	       AST_Decl(AST_Decl::NT_attr, n, p),
	       pd_readonly(ro)
{
}

/*
 * Private operations
 */

/*
 * Public operations
 */

/*
 * Redefinition of inherited virtual operations
 */

/*
 * Dump this AST_Attribute to the ostream o
 */
void
AST_Attribute::dump(ostream &o)
{
  o << (pd_readonly ? "readonly" : "") << " attribute ";
  AST_Field::dump(o);
}

/*
 * Data accessors
 */

const boolean
AST_Attribute::readonly()
{
  return pd_readonly;
}

/*
 * Narrowing methods
 */
IMPL_NARROW_METHODS1(AST_Attribute, AST_Field)
IMPL_NARROW_FROM_DECL(AST_Attribute)
