/*
 *======================================================================
 *
 *Copyright 1992 Sun Microsystems, Inc.
 *
 *The Interface Definition Language Compiler Front End (CFE) is made
 *available for use provided that this legend is included on all media and
 *documentation and as a part of the software program in whole or part.
 *Users may copy and extend functionality (but may not remove
 *functionality) of the Interface Definition Language CFE without charge,
 *but are not authorized to license or distribute it to anyone else except
 *as part of a product or program developed by the user or with the express
 *written consent of Sun Microsystems, Inc.
 *
 *The names of Sun Microsystems, Inc. and any of its subsidiaries may not
 *be used in advertising or publicity pertaining to distribution of
 *Interface Definition Language CFE as permitted herein.
 *
 *The Interface Definition Language CFE may not be exported outside the
 *United States without first obtaining the appropriate government
 *approvals.
 *
 *INTERFACE DEFINITION LANGUAGE CFE IS PROVIDED AS IS WITH NO WARRANTIES
 *OF ANY KIND INCLUDING THE WARRANTIES OF DESIGN, MERCHANTIBILITY AND
 *FITNESS FOR A PARTICULAR PURPOSE, NONINFRINGEMENT, OR ARISING FROM A
 *COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 *
 *Interface Definition Language CFE is provided with no support and
 *without any obligation on the part of Sun Microsystems, Inc. or any of
 *its subsidiaries or affiliates to assist in its use, correction,
 *modification or enhancement.
 *
 *SUN MICROSYSTEMS, INC. OR ANY OF ITS SUBSIDIARIES OR AFFILIATES SHALL
 *HAVE NO LIABILITY WITH RESPECT TO THE INFRINGEMENT OF COPYRIGHTS, TRADE
 *SECRETS OR ANY PATENTS BY INTERFACE DEFINITION LANGUAGE CFE OR ANY PART
 *THEREOF.
 *
 *IN NO EVENT WILL SUN MICROSYSTEMS, INC. OR ANY OF ITS SUBSIDIARIES OR
 *AFFILIATES BE LIABLE FOR ANY LOST REVENUE OR PROFITS OR OTHER SPECIAL,
 *INDIRECT AND CONSEQUENTIAL DAMAGES, EVEN IF SUN HAS BEEN ADVISED OF THE
 *POSSIBILITY OF SUCH DAMAGES.
 *
 *SunSoft, Inc.  
 *2550 Garcia Avenue 
 *Mountain View, California  94043
 *
 *
 *
 *
 *======================================================================
 */

#pragma ident "%@(#)fe_declarator.cc	1.30% %92/06/10% Sun Microsystems"

/*
 * fe_declarator.cc - Implementation of FE private declaration
 *		      statement class
 *
 * The FE_Declarator class is used to hold together complex declarations
 * until all the components are parsed and before they are composed into
 * an AST_Type node.
 */

#include	<idl.hh>
#include	<idl_extern.hh>

#include	<fe_private.hh>

/*
 * Constructor(s) and destructor
 */

FE_Declarator::FE_Declarator(UTL_ScopedName *n, DeclaratorType dt,
			     AST_Decl *cp)
	     : pd_name(n),
	       pd_complex_part(cp),
	       pd_decl_type(dt)
{
}

/*
 * Private operations
 */

/*
 * Public operations
 */

// Compose the type of the complex declarator (if any) with the base
// type supplied in ct
AST_Type *
FE_Declarator::compose(AST_Type *ct)
{
  AST_Array	*arr;

  if (pd_decl_type == FD_simple || pd_complex_part == NULL) {
    return ct;
  }
  if (pd_complex_part->node_type() == AST_Decl::NT_array) {
    arr = AST_Array::narrow_from_decl(pd_complex_part);
    arr->set_base_type(ct);
    return arr;
  }
}

/*
 * Redefinition of inherited virtual operations
 */

/*
 * Data accessors
 */

AST_Decl *
FE_Declarator::complex_part()
{
  return pd_complex_part;
}

UTL_ScopedName *
FE_Declarator::name()
{
  return pd_name;
}

FE_Declarator::DeclaratorType
FE_Declarator::decl_type()
{
  return pd_decl_type;
}
